from typing import Any, Dict, Optional

from dagster import check
from dagster.core.definitions import PipelineDefinition
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.utils.backcompat import experimental_fn_warning


def validate_run_config(
    pipeline_def: PipelineDefinition,
    run_config: Optional[Dict[str, Any]] = None,
    mode: Optional[str] = None,
) -> Dict[str, Any]:
    """Function to validate a provided run config blob against a given pipeline and mode.

    If validation is successful, this function will return a dictionary representation of the
    validated config actually used during execution.

    Args:
        pipeline_def (PipelineDefinition): The pipeline definition to validate run config against
        run_config (Optional[Dict[str, Any]]): The run config to validate
        mode (str): The mode of the pipeline to validate against (different modes may require
            different config)

    Returns:
        Dict[str, Any]: A dictionary representation of the validated config.
    """

    experimental_fn_warning("validate_run_config")

    pipeline_def = check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)
    run_config = check.opt_dict_param(run_config, "run_config", key_type=str)
    mode = check.opt_str_param(mode, "mode", default=pipeline_def.get_default_mode_name())

    return EnvironmentConfig.build(pipeline_def, run_config, mode=mode).to_dict()



from collections import namedtuple
from enum import Enum

from dagster import check
from dagster.core.definitions.run_request import JobType
from dagster.core.host_representation.origin import ExternalJobOrigin
from dagster.serdes import whitelist_for_serdes
from dagster.utils import merge_dicts
from dagster.utils.error import SerializableErrorInfo


@whitelist_for_serdes
class JobStatus(Enum):
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"


@whitelist_for_serdes
class SensorJobData(
    namedtuple("_SensorJobData", "last_tick_timestamp last_run_key min_interval cursor")
):
    def __new__(cls, last_tick_timestamp=None, last_run_key=None, min_interval=None, cursor=None):
        return super(SensorJobData, cls).__new__(
            cls,
            check.opt_float_param(last_tick_timestamp, "last_tick_timestamp"),
            check.opt_str_param(last_run_key, "last_run_key"),
            check.opt_int_param(min_interval, "min_interval"),
            check.opt_str_param(cursor, "cursor"),
        )


@whitelist_for_serdes
class ScheduleJobData(namedtuple("_ScheduleJobData", "cron_schedule start_timestamp scheduler")):
    def __new__(cls, cron_schedule, start_timestamp=None, scheduler=None):
        return super(ScheduleJobData, cls).__new__(
            cls,
            check.str_param(cron_schedule, "cron_schedule"),
            # Time in UTC at which the user started running the schedule (distinct from
            # `start_date` on partition-based schedules, which is used to define
            # the range of partitions)
            check.opt_float_param(start_timestamp, "start_timestamp"),
            check.opt_str_param(scheduler, "scheduler"),
        )


def check_job_data(job_type, job_specific_data):
    check.inst_param(job_type, "job_type", JobType)
    if job_type == JobType.SCHEDULE:
        check.inst_param(job_specific_data, "job_specific_data", ScheduleJobData)
    elif job_type == JobType.SENSOR:
        check.opt_inst_param(job_specific_data, "job_specific_data", SensorJobData)
    else:
        check.failed(
            "Unexpected job type {}, expected one of JobType.SENSOR, JobType.SCHEDULE".format(
                job_type
            )
        )

    return job_specific_data


@whitelist_for_serdes
class JobState(namedtuple("_JobState", "origin job_type status job_specific_data")):
    def __new__(cls, origin, job_type, status, job_specific_data=None):
        return super(JobState, cls).__new__(
            cls,
            check.inst_param(origin, "origin", ExternalJobOrigin),
            check.inst_param(job_type, "job_type", JobType),
            check.inst_param(status, "status", JobStatus),
            check_job_data(job_type, job_specific_data),
        )

    @property
    def name(self):
        return self.origin.job_name

    @property
    def job_name(self):
        return self.origin.job_name

    @property
    def repository_origin_id(self):
        return self.origin.external_repository_origin.get_id()

    @property
    def job_origin_id(self):
        return self.origin.get_id()

    def with_status(self, status):
        check.inst_param(status, "status", JobStatus)
        return JobState(
            self.origin,
            job_type=self.job_type,
            status=status,
            job_specific_data=self.job_specific_data,
        )

    def with_data(self, job_specific_data):
        check_job_data(self.job_type, job_specific_data)
        return JobState(
            self.origin,
            job_type=self.job_type,
            status=self.status,
            job_specific_data=job_specific_data,
        )


@whitelist_for_serdes
class JobTickStatus(Enum):
    STARTED = "STARTED"
    SKIPPED = "SKIPPED"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


@whitelist_for_serdes
class JobTick(namedtuple("_JobTick", "tick_id job_tick_data")):
    def __new__(cls, tick_id, job_tick_data):
        return super(JobTick, cls).__new__(
            cls,
            check.int_param(tick_id, "tick_id"),
            check.inst_param(job_tick_data, "job_tick_data", JobTickData),
        )

    def with_status(self, status, **kwargs):
        check.inst_param(status, "status", JobTickStatus)
        return self._replace(job_tick_data=self.job_tick_data.with_status(status, **kwargs))

    def with_reason(self, skip_reason):
        check.opt_str_param(skip_reason, "skip_reason")
        return self._replace(job_tick_data=self.job_tick_data.with_reason(skip_reason))

    def with_run(self, run_id, run_key=None):
        return self._replace(job_tick_data=self.job_tick_data.with_run(run_id, run_key))

    def with_cursor(self, cursor):
        return self._replace(job_tick_data=self.job_tick_data.with_cursor(cursor))

    @property
    def job_origin_id(self):
        return self.job_tick_data.job_origin_id

    @property
    def job_name(self):
        return self.job_tick_data.job_name

    @property
    def job_type(self):
        return self.job_tick_data.job_type

    @property
    def timestamp(self):
        return self.job_tick_data.timestamp

    @property
    def status(self):
        return self.job_tick_data.status

    @property
    def run_ids(self):
        return self.job_tick_data.run_ids

    @property
    def run_keys(self):
        return self.job_tick_data.run_keys

    @property
    def error(self):
        return self.job_tick_data.error

    @property
    def skip_reason(self):
        return self.job_tick_data.skip_reason

    @property
    def cursor(self):
        return self.job_tick_data.cursor


@whitelist_for_serdes
class JobTickData(
    namedtuple(
        "_JobTickData",
        "job_origin_id job_name job_type status timestamp run_ids run_keys error skip_reason cursor",
    )
):
    def __new__(
        cls,
        job_origin_id,
        job_name,
        job_type,
        status,
        timestamp,
        run_ids=None,
        run_keys=None,
        error=None,
        skip_reason=None,
        cursor=None,
    ):
        """
        This class defines the data that is serialized and stored in ``JobStorage``. We depend
        on the job storage implementation to provide job tick ids, and therefore
        separate all other data into this serializable class that can be stored independently of the
        id

        Arguments:
            job_origin_id (str): The id of the job target for this tick
            job_name (str): The name of the job for this tick
            job_type (JobType): The type of this job for this tick
            status (JobTickStatus): The status of the tick, which can be updated
            timestamp (float): The timestamp at which this job evaluation started

        Keyword Arguments:
            run_id (str): The run created by the tick.
            error (SerializableErrorInfo): The error caught during job execution. This is set
                only when the status is ``JobTickStatus.Failure``
            skip_reason (str): message for why the tick was skipped
        """

        _validate_job_tick_args(job_type, status, run_ids, error, skip_reason)
        return super(JobTickData, cls).__new__(
            cls,
            check.str_param(job_origin_id, "job_origin_id"),
            check.str_param(job_name, "job_name"),
            check.inst_param(job_type, "job_type", JobType),
            check.inst_param(status, "status", JobTickStatus),
            check.float_param(timestamp, "timestamp"),
            check.opt_list_param(run_ids, "run_ids", of_type=str),
            check.opt_list_param(run_keys, "run_keys", of_type=str),
            error,  # validated in _validate_job_tick_args
            skip_reason,  # validated in _validate_job_tick_args
            cursor=check.opt_str_param(cursor, "cursor"),
        )

    def with_status(self, status, error=None, timestamp=None):
        return JobTickData(
            **merge_dicts(
                self._asdict(),
                {
                    "status": status,
                    "timestamp": timestamp if timestamp is not None else self.timestamp,
                    "error": error if error is not None else self.error,
                },
            )
        )

    def with_run(self, run_id, run_key=None):
        check.str_param(run_id, "run_id")
        return JobTickData(
            **merge_dicts(
                self._asdict(),
                {
                    "run_ids": [*self.run_ids, run_id],
                    "run_keys": [*self.run_keys, run_key] if run_key else self.run_keys,
                },
            )
        )

    def with_reason(self, skip_reason):
        return JobTickData(
            **merge_dicts(
                self._asdict(), {"skip_reason": check.opt_str_param(skip_reason, "skip_reason")}
            )
        )

    def with_cursor(self, cursor):
        return JobTickData(
            **merge_dicts(self._asdict(), {"cursor": check.opt_str_param(cursor, "cursor")})
        )


def _validate_job_tick_args(job_type, status, run_ids=None, error=None, skip_reason=None):
    check.inst_param(job_type, "job_type", JobType)
    check.inst_param(status, "status", JobTickStatus)

    if status == JobTickStatus.SUCCESS:
        check.list_param(run_ids, "run_ids", of_type=str)
        check.invariant(error is None, desc="Job tick status is SUCCESS, but error was provided")
    elif status == JobTickStatus.FAILURE:
        check.inst_param(error, "error", SerializableErrorInfo)
    else:
        check.invariant(error is None, "Job tick status was not FAILURE but error was provided")

    if skip_reason:
        check.invariant(
            status == JobTickStatus.SKIPPED,
            "Job tick status was not SKIPPED but skip_reason was provided",
        )


class JobTickStatsSnapshot(
    namedtuple(
        "JobTickStatsSnapshot",
        ("ticks_started ticks_succeeded ticks_skipped ticks_failed"),
    )
):
    def __new__(
        cls,
        ticks_started,
        ticks_succeeded,
        ticks_skipped,
        ticks_failed,
    ):
        return super(JobTickStatsSnapshot, cls).__new__(
            cls,
            ticks_started=check.int_param(ticks_started, "ticks_started"),
            ticks_succeeded=check.int_param(ticks_succeeded, "ticks_succeeded"),
            ticks_skipped=check.int_param(ticks_skipped, "ticks_skipped"),
            ticks_failed=check.int_param(ticks_failed, "ticks_failed"),
        )



from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Union

from dagster import check
from dagster.cli.workspace.workspace import (
    Workspace,
    WorkspaceLocationLoadStatus,
    WorkspaceSnapshot,
)
from dagster.core.execution.plan.state import KnownExecutionState
from dagster.core.host_representation import (
    ExternalExecutionPlan,
    ExternalPipeline,
    PipelineSelector,
    RepositoryHandle,
    RepositoryLocation,
)
from dagster.core.host_representation.grpc_server_state_subscriber import (
    LocationStateChangeEvent,
    LocationStateChangeEventType,
    LocationStateSubscriber,
)
from dagster.core.instance import DagsterInstance

if TYPE_CHECKING:
    from rx.subjects import Subject
    from dagster.core.host_representation import (
        ExternalPartitionSetExecutionParamData,
        ExternalPartitionExecutionErrorData,
        ExternalPartitionNamesData,
        ExternalPartitionConfigData,
        ExternalPartitionTagsData,
    )
    from dagster.utils.error import SerializableErrorInfo


class WorkspaceRequestContext(NamedTuple):
    """
    This class is request-scoped object that stores (1) a reference to all repository locations
    that exist on the `WorkspaceProcessContext` at the start of the request and (2) a snapshot of the
    `Workspace` at the start of the request.

    This object is needed because a process context and the repository locations on that context can
    be updated (for example, from a thread on the process context). If a request is accessing a
    repository location at the same time the repository location was being cleaned up, we would run
    into errors.
    """

    instance: DagsterInstance
    workspace_snapshot: WorkspaceSnapshot
    process_context: "WorkspaceProcessContext"
    version: Optional[str] = None

    @property
    def repository_locations_dict(self) -> Dict[str, RepositoryLocation]:
        return self.workspace_snapshot.repository_locations_dict

    @property
    def repository_locations(self) -> List[RepositoryLocation]:
        return list(self.repository_locations_dict.values())

    @property
    def repository_location_names(self) -> List[str]:
        return self.workspace_snapshot.repository_location_names

    @property
    def read_only(self) -> bool:
        return self.process_context.read_only

    def repository_location_errors(self) -> List["SerializableErrorInfo"]:
        return self.workspace_snapshot.repository_location_errors

    def get_repository_location(self, name: str) -> RepositoryLocation:
        return self.repository_locations_dict[name]

    def get_load_status(self, name: str) -> WorkspaceLocationLoadStatus:
        return self.workspace_snapshot.get_load_status(name)

    def has_repository_location_error(self, name: str) -> bool:
        return self.workspace_snapshot.has_repository_location_error(name)

    def get_repository_location_error(self, name: str) -> "SerializableErrorInfo":
        return self.workspace_snapshot.get_repository_location_error(name)

    def has_repository_location(self, name: str) -> bool:
        return name in self.repository_locations_dict

    def is_reload_supported(self, name: str) -> bool:
        return self.workspace_snapshot.is_reload_supported(name)

    def reload_repository_location(self, name: str) -> "WorkspaceRequestContext":
        # This method reloads the location on the process context, and returns a new
        # request context created from the updated process context
        updated_process_context = self.process_context.reload_repository_location(name)
        return updated_process_context.create_request_context()

    def reload_workspace(self) -> "WorkspaceRequestContext":
        self.process_context.reload_workspace()
        return self.process_context.create_request_context()

    def has_external_pipeline(self, selector: PipelineSelector) -> bool:
        check.inst_param(selector, "selector", PipelineSelector)
        loc = self.repository_locations_dict.get(selector.location_name)
        return (
            loc is not None
            and loc.has_repository(selector.repository_name)
            and loc.get_repository(selector.repository_name).has_external_pipeline(
                selector.pipeline_name
            )
        )

    def get_full_external_pipeline(self, selector: PipelineSelector) -> ExternalPipeline:
        return (
            self.repository_locations_dict[selector.location_name]
            .get_repository(selector.repository_name)
            .get_full_external_pipeline(selector.pipeline_name)
        )

    def get_external_execution_plan(
        self,
        external_pipeline: ExternalPipeline,
        run_config: dict,
        mode: str,
        step_keys_to_execute: List[str],
        known_state: KnownExecutionState,
    ) -> ExternalExecutionPlan:
        return self.repository_locations_dict[
            external_pipeline.handle.location_name
        ].get_external_execution_plan(
            external_pipeline=external_pipeline,
            run_config=run_config,
            mode=mode,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
        )

    def get_external_partition_config(
        self, repository_handle: RepositoryHandle, partition_set_name: str, partition_name: str
    ) -> Union["ExternalPartitionConfigData", "ExternalPartitionExecutionErrorData"]:
        return self.repository_locations_dict[
            repository_handle.repository_location.name
        ].get_external_partition_config(
            repository_handle=repository_handle,
            partition_set_name=partition_set_name,
            partition_name=partition_name,
        )

    def get_external_partition_tags(
        self, repository_handle: RepositoryHandle, partition_set_name: str, partition_name: str
    ) -> Union["ExternalPartitionTagsData", "ExternalPartitionExecutionErrorData"]:
        return self.repository_locations_dict[
            repository_handle.repository_location.name
        ].get_external_partition_tags(
            repository_handle=repository_handle,
            partition_set_name=partition_set_name,
            partition_name=partition_name,
        )

    def get_external_partition_names(
        self, repository_handle: RepositoryHandle, partition_set_name: str
    ) -> Union["ExternalPartitionNamesData", "ExternalPartitionExecutionErrorData"]:
        return self.repository_locations_dict[
            repository_handle.repository_location.name
        ].get_external_partition_names(repository_handle, partition_set_name)

    def get_external_partition_set_execution_param_data(
        self,
        repository_handle: RepositoryHandle,
        partition_set_name: str,
        partition_names: List[str],
    ) -> Union["ExternalPartitionSetExecutionParamData", "ExternalPartitionExecutionErrorData"]:
        return self.repository_locations_dict[
            repository_handle.repository_location.name
        ].get_external_partition_set_execution_param_data(
            repository_handle=repository_handle,
            partition_set_name=partition_set_name,
            partition_names=partition_names,
        )


class WorkspaceProcessContext:
    """
    This class is process-scoped object that is initialized using the repository handles from a
    Workspace. The responsibility of this class is to:

    1. Maintain an update-to-date dictionary of repository locations
    1. Create `WorkspaceRequestContexts` whever a request is made
    2. Run watch thread processes that monitor repository locations

    In most cases, you will want to create a `WorkspaceRequestContext` to make use of this class.
    """

    def __init__(
        self,
        instance: DagsterInstance,
        workspace: Workspace,
        version: Optional[str] = None,
        read_only: bool = False,
    ):

        check.inst_param(instance, "instance", DagsterInstance)
        check.inst_param(workspace, "workspace", Workspace)
        check.opt_str_param(version, "version")
        check.bool_param(read_only, "read_only")

        # lazy import for perf
        from rx.subjects import Subject

        self._instance = check.inst_param(instance, "instance", DagsterInstance)
        self._workspace = workspace

        self._location_state_events = Subject()
        self._location_state_subscriber = LocationStateSubscriber(
            self._location_state_events_handler
        )

        self.read_only = read_only
        self.version = version
        self.set_state_subscribers()

    def set_state_subscribers(self):
        self._workspace.add_state_subscriber(self._location_state_subscriber)

    def create_request_context(self) -> WorkspaceRequestContext:
        return WorkspaceRequestContext(
            instance=self.instance,
            workspace_snapshot=self._workspace.create_snapshot(),
            process_context=self,
            version=self.version,
        )

    @property
    def instance(self) -> DagsterInstance:
        return self._instance

    @property
    def workspace(self) -> Workspace:
        return self._workspace

    @property
    def repository_locations(self) -> List[RepositoryLocation]:
        return self._workspace.repository_locations

    @property
    def location_state_events(self) -> "Subject":
        return self._location_state_events

    def _location_state_events_handler(self, event: LocationStateChangeEvent) -> None:
        # If the server was updated or we were not able to reconnect, we immediately reload the
        # location handle
        if event.event_type in (
            LocationStateChangeEventType.LOCATION_UPDATED,
            LocationStateChangeEventType.LOCATION_ERROR,
        ):
            # In case of an updated location, reload the handle to get updated repository data and
            # re-attach a subscriber
            # In case of a location error, just reload the handle in order to update the workspace
            # with the correct error messages
            self.reload_repository_location(event.location_name)

        self._location_state_events.on_next(event)

    def reload_repository_location(self, name: str) -> "WorkspaceProcessContext":
        self._workspace.reload_repository_location(name)
        return self

    def reload_workspace(self) -> None:
        self._workspace.reload_workspace()
        self.set_state_subscribers()



from dagster import solid


@solid
def hello(_context):
    """
    A solid definition. This example solid outputs a single string.

    For more hints about writing Dagster solids, see our documentation overview on Solids:
    https://docs.dagster.io/overview/solids-pipelines/solids
    """
    return "Hello, Dagster!"
